//Title: MediaPlayerNode Class
//Platform: BeOS 5
//Version: 2002.03.19
//Description:
//	A class that creates a BBufferProducer Media Node, and plays media from a BMediaFile.
//
	
//Copyright (C) 2001, 2002 Kevin H. Patterson
//Available for use under the terms of the BSD license agreement:
//See the associated file "License.txt" for details.

#include "MediaPlayerNode.h"

#include <Entry.h>
#include <Autolock.h>

#include <iostream.h>
#include "Log.h"

MediaPlayerNode::MediaPlayerNode( const char *uMediaFileName )
:
	BMediaNode( "KHP-MediaPlayerNode" ),
	BBufferProducer( B_MEDIA_UNKNOWN_TYPE ),
	BMediaEventLooper(),
	iInitCheck( B_ERROR ),
	iMediaFile( NULL ),
	iMediaStreamCount( 0 ),
	iMediaStream(),
	iRunning( false ),
	iNodeLock( new BLocker() )
{
	Log("MediaPlayerNode::MediaPlayerNode()\n");
	entry_ref FileEntryRef;
	if (get_ref_for_path( uMediaFileName, &FileEntryRef ) < B_OK)
		return;
	
	iMediaFile = new BMediaFile( &FileEntryRef );
	if (iMediaFile == NULL)
		return;
	if (iMediaFile->InitCheck() < B_OK)
		return;

	bool tDeleteInstance;

	for (int32 TrackIndex( 0 ); TrackIndex < iMediaFile->CountTracks(); ++TrackIndex) {
		iMediaStream[iMediaStreamCount] = new OutputStream( this, iMediaFile, TrackIndex );
		if (iMediaStream[iMediaStreamCount] != NULL) {
			tDeleteInstance = true;
			if (iMediaStream[iMediaStreamCount]->InitCheck() >= B_OK) {
				if (
						iMediaStream[iMediaStreamCount]->IsVideo()
					|| iMediaStream[iMediaStreamCount]->IsAudio()
				)
				{
					iMediaStream[iMediaStreamCount]->Output()->source.id = iMediaStreamCount;
					tDeleteInstance = false;
					iMediaStreamCount++;
				}
			}
			if (tDeleteInstance) {
				delete iMediaStream[iMediaStreamCount];
				iMediaStream[iMediaStreamCount] = NULL;
			}
		}
	}

	if (iMediaStreamCount < 1) {
		cout << "ERROR: No decodable tracks found!" << endl;
		return;
	}
	
	iInitCheck = B_OK;
}


MediaPlayerNode::~MediaPlayerNode() {
	Log("MediaPlayerNode::~MediaPlayerNode()\n");
	BMediaEventLooper::Quit();
	if (iInitCheck >= B_OK) {
		if (iRunning)
			HandleStop();
	}
	for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex)
		delete iMediaStream[tIndex];
	delete iMediaFile;
	delete iNodeLock;
}


status_t MediaPlayerNode::InitCheck() const {
	return iInitCheck;
}


// BMediaNode

port_id MediaPlayerNode::ControlPort() const {
	return BMediaNode::ControlPort();
}


BMediaAddOn * MediaPlayerNode::AddOn( int32 * outInternalID ) const {
	return NULL;
}


status_t MediaPlayerNode::HandleMessage(
	int32 message,
	const void * data,
	size_t size )
{
	//We don't use any private message protocols
	status_t tStatus;
	tStatus = BBufferProducer::HandleMessage( message, data, size );
	if (tStatus < B_OK)
		tStatus = BMediaNode::HandleMessage( message, data, size );
	if (tStatus < B_OK) {
		BMediaNode::HandleBadMessage( message, data, size );
		return B_ERROR;
	}
	return tStatus;
}


void MediaPlayerNode::Preroll() {
	Log("MediaPlayerNode::Preroll()\n");
	for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex)
		iMediaStream[tIndex]->Preroll();
	BMediaEventLooper::Preroll();
}


void MediaPlayerNode::SetTimeSource( BTimeSource * time_source ) {
	Log("MediaPlayerNode::SetTimeSource()\n");
	BMediaNode::SetTimeSource( time_source );
}


status_t MediaPlayerNode::RequestCompleted( const media_request_info & info ) {
	Log("MediaPlayerNode::RequestCompleted()\n");
	return BMediaNode::RequestCompleted( info );
}


// BBufferProducer

status_t MediaPlayerNode::FormatSuggestionRequested(
	media_type type,
	int32 quality,
	media_format * format )
{
	Log("MediaPlayerNode::FormatSuggestionRequested()\n");
	if (format == NULL)
		return B_BAD_VALUE;
	if (type == B_MEDIA_RAW_AUDIO) {
		for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex) {
			if (iMediaStream[tIndex]->IsAudio()) {
				*format = *iMediaStream[tIndex]->PreferredFormat();
				return B_OK;
			}
		}
	}
	if (type == B_MEDIA_RAW_VIDEO) {
		for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex) {
			if (iMediaStream[tIndex]->IsVideo()) {
				*format = *iMediaStream[tIndex]->PreferredFormat();
				return B_OK;
			}
		}
	}
	return B_MEDIA_BAD_FORMAT;
}


status_t MediaPlayerNode::FormatProposal(
	const media_source & output,
	media_format * format )
{
	Log("MediaPlayerNode::FormatProposal()\n");
	if (format == NULL)
		return B_BAD_VALUE;
	if ((output.id < 0) || (output.id >= iMediaStreamCount))
		return B_MEDIA_BAD_SOURCE;
	if (output != iMediaStream[output.id]->Output()->source)
		return B_MEDIA_BAD_SOURCE;
	return iMediaStream[output.id]->SpecializeFormat( format );
}


status_t MediaPlayerNode::FormatChangeRequested(
	const media_source & source,
	const media_destination & destination,
	media_format * format,
	int32 * /* _deprecated_ */ )
{
	Log("MediaPlayerNode::FormatChangeRequested()\n");
	return B_ERROR;
}


status_t MediaPlayerNode::VideoClippingChanged(
	const media_source & source,
	int16 num_shorts,
	int16 * clip_data,
	const media_video_display_info & display,
	int32 * _deprecated_ )
{
	Log("MediaPlayerNode::VideoClippingChanged()\n");
	return B_ERROR;
}


status_t MediaPlayerNode::GetNextOutput(
	int32 *cookie,
	media_output *outOutput )
{
	Log("MediaPlayerNode::GetNextOutput()\n");
	if (outOutput == NULL)
		return B_BAD_VALUE;
	if ((*cookie < 0) || (*cookie >= iMediaStreamCount))
		return B_BAD_INDEX;
	*outOutput = *iMediaStream[*cookie]->Output();
	*cookie += 1;
	return B_OK;
}


status_t MediaPlayerNode::DisposeOutputCookie( int32 cookie ) {
	return B_OK;
}


status_t MediaPlayerNode::SetBufferGroup(
	const media_source & output,
	BBufferGroup * group )
{
	Log("MediaPlayerNode::SetBufferGroup()\n");
	if ((output.id < 0) || (output.id >= iMediaStreamCount))
		return B_MEDIA_BAD_SOURCE;
	if (output != iMediaStream[output.id]->Output()->source)
		return B_MEDIA_BAD_SOURCE;
	iMediaStream[output.id]->SetBufferGroup( group );
	return B_OK;
}


status_t MediaPlayerNode::GetLatency( bigtime_t * latency ) {
	*latency = EventLatency() + SchedulingLatency();
	return B_OK;
}


status_t MediaPlayerNode::PrepareToConnect(
	const media_source & output,
	const media_destination & where,
	media_format * format,
	media_source * outSource,
	char * outName )
{
	Log("MediaPlayerNode::PrepareToConnect()\n");
	if ((output.id < 0) || (output.id >= iMediaStreamCount))
		return B_MEDIA_BAD_SOURCE;
	if (output != iMediaStream[output.id]->Output()->source)
		return B_MEDIA_BAD_SOURCE;
	if (iMediaStream[output.id]->Output()->destination != media_destination::null)
		return B_MEDIA_ALREADY_CONNECTED;
	if ((outSource == NULL) || (outName == NULL))
		return B_BAD_VALUE;
	*outSource = iMediaStream[output.id]->Output()->source;
	strncpy( outName, iMediaStream[output.id]->Output()->name, B_MEDIA_NAME_LENGTH );
	return iMediaStream[output.id]->SpecializeFormat( format );
}


void MediaPlayerNode::Connect(
	status_t error,
	const media_source & output,
	const media_destination & destination,
	const media_format & format,
	char * outName )
{
	Log("MediaPlayerNode::Connect()\n");
	if ((output.id < 0) || (output.id >= iMediaStreamCount)) {
		error = B_MEDIA_BAD_SOURCE;
		return;
	}
	if (output != iMediaStream[output.id]->Output()->source) {
		error = B_MEDIA_BAD_SOURCE;
		return;
	}
	
	iMediaStream[output.id]->Connect( error, destination, format, outName );

	//Set the buffer duration
	{
		int32 numBuffers;
		iMediaStream[output.id]->BufferGroup()->CountBuffers( &numBuffers );
		SetBufferDuration( (1000000LL * numBuffers) / iMediaStream[output.id]->Output()->format.u.raw_video.field_rate );
	}
	
	//Latency
	{
		//Determine downstream latency
		media_node_id nodeID;
		bigtime_t tDownstreamLatency;
		FindLatencyFor( iMediaStream[output.id]->Output()->destination, &tDownstreamLatency, &nodeID );
		iMediaStream[output.id]->SetDownstreamLatency( tDownstreamLatency );
		
		//Set initial latency
		SetInitialLatency( iMediaStream[output.id]->GetLatency() );
	}
	RecomputeLatency();
}


void MediaPlayerNode::Disconnect(
	const media_source & output,
	const media_destination & where )
{
	Log("MediaPlayerNode::Disconnect()\n");
	if ((output.id < 0) || (output.id >= iMediaStreamCount))
		return;
	if (output != iMediaStream[output.id]->Output()->source)
		return;
	iMediaStream[output.id]->Disconnect( where );
}


void MediaPlayerNode::LateNoticeReceived(
	const media_source & output,
	bigtime_t how_much,
	bigtime_t performance_time )
{
	Log("MediaPlayerNode::LateNoticeReceived()\n");
	if ((output.id < 0) || (output.id >= iMediaStreamCount))
		return;
	if (output != iMediaStream[output.id]->Output()->source)
		return;
		
	//Handle the late notice according to our current Run Mode
	switch (RunMode()) {
	case B_RECORDING:
		//Don't do anything, since hardware capture can't be adjusted
		break;
	default:
	case B_INCREASE_LATENCY:
		//Increase our latency
		iMediaStream[output.id]->SetExtraLatency( how_much );
		//Skip buffers?
		//BETA: Needs implementation
		break;
	}
}


void MediaPlayerNode::EnableOutput(
	const media_source & output,
	bool enabled,
	int32 * _deprecated_ )
{
	Log("MediaPlayerNode::EnableOutput()\n");
	if ((output.id < 0) || (output.id >= iMediaStreamCount))
		return;
	if (output != iMediaStream[output.id]->Output()->source)
		return;
	iMediaStream[output.id]->EnableOutput( enabled );
}


status_t MediaPlayerNode::SetPlayRate( int32 numer, int32 denom ) {
	Log("MediaPlayerNode::SetPlayRate()\n");
	return B_ERROR;
}


void MediaPlayerNode::AdditionalBufferRequested(
	const media_source & source,
	media_buffer_id prev_buffer,
	bigtime_t prev_time,
	const media_seek_tag * prev_tag )
{
	Log("MediaPlayerNode::AdditionalBufferRequested()\n");
	//BETA: Needs implementation for Offline Mode!
	BBufferProducer::AdditionalBufferRequested(
		source, prev_buffer, prev_time, prev_tag );
}


void MediaPlayerNode::LatencyChanged(
	const media_source & output,
	const media_destination & destination,
	bigtime_t new_latency,
	uint32 flags )
{
	if ((output.id < 0) || (output.id >= iMediaStreamCount))
		return;
	if (output != iMediaStream[output.id]->Output()->source)
		return;
	if (destination != iMediaStream[output.id]->Output()->destination) {
		iMediaStream[output.id]->SetDownstreamLatency( new_latency );
	}
}


// BMediaEventLooper

void MediaPlayerNode::NodeRegistered() {
	Log("MediaPlayerNode::NodeRegistered()\n");
	if (iInitCheck < B_OK)
		ReportError( B_NODE_IN_DISTRESS );
	else {
		for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex) {
			iMediaStream[tIndex]->Output()->node = Node();
			iMediaStream[tIndex]->Output()->source.port = ControlPort();
			iMediaStream[tIndex]->Output()->source.id = tIndex;
			strncpy( iMediaStream[tIndex]->Output()->name, Name(), B_MEDIA_NAME_LENGTH );
		}
	}
	BMediaEventLooper::NodeRegistered();
}


void MediaPlayerNode::Start( bigtime_t performance_time ) {
	Log("MediaPlayerNode::Start()\n");
	BMediaEventLooper::Start( performance_time );
}


void MediaPlayerNode::Stop( bigtime_t performance_time, bool immediate ) {
	Log("MediaPlayerNode::Stop()\n");
	BMediaEventLooper::Stop( performance_time, immediate );
}


void MediaPlayerNode::Seek( bigtime_t media_time, bigtime_t performance_time ) {
	Log("MediaPlayerNode::Seek()\n");
	BMediaEventLooper::Seek( media_time, performance_time );
}


void MediaPlayerNode::TimeWarp( bigtime_t at_real_time, bigtime_t to_performance_time ) {
	Log("MediaPlayerNode::TimeWarp()\n");
	BMediaEventLooper::TimeWarp( at_real_time, to_performance_time );
}


status_t MediaPlayerNode::AddTimer( bigtime_t at_performance_time, int32 cookie ) {
	Log("MediaPlayerNode::AddTimer()\n");
	return BMediaEventLooper::AddTimer( at_performance_time, cookie );
}


void MediaPlayerNode::SetRunMode( run_mode mode ) {
	Log("MediaPlayerNode::SetRunMode()\n");
	switch (mode) {
	//BETA: If we can't support a particular mode, this is the place to report an error!
	//ReportError( B_NODE_FAILED_SET_RUN_MODE );
	default:
		BMediaEventLooper::SetRunMode( mode );
	}
}


void MediaPlayerNode::HandleEvent(
	const media_timed_event *event,
	bigtime_t lateness,
	bool realTimeEvent = false )
{
	Log("MediaPlayerNode::HandleEvent()\n");
	switch (event->type) {
	case BTimedEventQueue::B_START:
		HandleStart( event->event_time );
		break;
	case BTimedEventQueue::B_STOP:
		HandleStop( event->event_time );
		break;
	case BTimedEventQueue::B_SEEK:
		HandleSeek( event->bigdata, event->event_time );
		break;
	case BTimedEventQueue::B_WARP:
		HandleTimeWarp( event->bigdata );
		break;
	case BTimedEventQueue::B_HANDLE_BUFFER:
	case BTimedEventQueue::B_DATA_STATUS:
	case BTimedEventQueue::B_PARAMETER:
	default:		
		//BETA: Unhandled event...
		break;
	}
}


void MediaPlayerNode::CleanUpEvent( const media_timed_event *event ) {
	BMediaEventLooper::CleanUpEvent( event );
}


bigtime_t MediaPlayerNode::OfflineTime() {
	return BMediaEventLooper::OfflineTime();
}


void MediaPlayerNode::ControlLoop() {
	BMediaEventLooper::ControlLoop();
}


status_t MediaPlayerNode::DeleteHook( BMediaNode * node ) {
	return BMediaEventLooper::DeleteHook( node );
}


//Implementation

void MediaPlayerNode::HandleStart( bigtime_t performance_time ) {
	if (iRunning) {
		Log("HandleStart: Node already started\n");
		return;
	}
	for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex)
		iMediaStream[tIndex]->Start( performance_time );
	iRunning = true;
}


void MediaPlayerNode::HandleStop()
{
	Log("HandleStop( immediate )\n");
	if (!iRunning) {
		Log("HandleStop: Node isn't running\n");
		return;
	}
	for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex)
			iMediaStream[tIndex]->Stop();
	EventQueue()->FlushEvents(
		0,
		BTimedEventQueue::B_ALWAYS,
		true,
		BTimedEventQueue::B_HANDLE_BUFFER );
	iRunning = false;
}


void MediaPlayerNode::HandleStop(	bigtime_t performance_time )
{
	if (!iRunning)
		return;
	for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex)
		iMediaStream[tIndex]->Stop( performance_time );
	EventQueue()->FlushEvents(
		0,
		BTimedEventQueue::B_ALWAYS,
		true,
		BTimedEventQueue::B_HANDLE_BUFFER );
	iRunning = false;
}


void MediaPlayerNode::HandleTimeWarp( bigtime_t performance_time ) {
	for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex)
		iMediaStream[tIndex]->TimeWarp( performance_time );
}


void MediaPlayerNode::HandleSeek( bigtime_t media_time, bigtime_t performance_time ) {
	bigtime_t tPerformanceTime( performance_time );
	if (tPerformanceTime == 0) {
		GetLatency( &tPerformanceTime );
		tPerformanceTime += TimeSource()->Now();
	}
	for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex)
		iMediaStream[tIndex]->Seek( media_time, tPerformanceTime );
}


status_t MediaPlayerNode::iSendBuffer(
	BBuffer * buffer,
	const media_destination & destination )
{
	BAutolock tAutoLock( iNodeLock );
	if (tAutoLock.IsLocked())
		return SendBuffer( buffer, destination );
	else
		return B_ERROR;
}


void MediaPlayerNode::iOutputLatencyChanged( bigtime_t new_latency ) {
	if (new_latency > EventLatency())
		SetEventLatency( new_latency );
}


bigtime_t MediaPlayerNode::RecomputeLatency() {
	bigtime_t MaxOutputLatency( 0 );
	bigtime_t tLatency;
	for (int32 tIndex( 0 ); tIndex < iMediaStreamCount; ++tIndex) {
		tLatency = iMediaStream[tIndex]->GetLatency();
		if (tLatency > MaxOutputLatency)
			MaxOutputLatency = tLatency;
	}
	SetEventLatency( MaxOutputLatency );
	return MaxOutputLatency;
}


bigtime_t MediaPlayerNode::Duration() const {
	if (iMediaStreamCount < 1)
		return -1;
	return iMediaStream[0]->Duration();
}


bigtime_t MediaPlayerNode::Position() const {
	if (iMediaStreamCount < 1)
		return 0;
	return iMediaStream[0]->Position();
}
