//Title: MK MediaPlayer Main Window Class
//Platform: BeOS 5
//Version: 2002.03.19
//Description:
//	A class that creates the application's "main" window.
//	This class inherits from BWindow, and runs a message loop for the window.
	
//Copyright (C) 2001, 2002 Kevin H. Patterson
//Available for use under the terms of the BSD license agreement:
//See the associated file "License.txt" for details.

#include "clsMainWindow.h"

#include <Button.h>
#include <Message.h>

#include "MediaPlayerNode.h"

#include <Menu.h>
#include <MenuItem.h>
#include <Entry.h>
#include <storage/Path.h>

#include <iostream.h>
#include "Log.h"

clsMainWindow::clsMainWindow( const char *uWindowTitle )
:
	BWindow(
		BRect( 64, 64, 320, 256 ),
		uWindowTitle,
		B_TITLED_WINDOW,
		0	),
	
	iMessenger( new BMessenger(NULL, this, NULL) ),
	iOpenFilePanel( NULL ),
	iMenuBar( NULL ),
	iSeekSlider( NULL ),
	iUpdateTimer( NULL ),
	FileName(""),
	
	iMediaPlayerNode( NULL ),
	iTimeSourceNode( NULL ),
	MediaProducer( media_node::null ),
	AudioConsumer( media_node::null ),
	VideoConsumer( media_node::null ),
	TimeSource( media_node::null ),
	ConnectionCount( 0 ),
	Connection(),
	UsesAudio( false ),
	UsesVideo( false )	
{
	if (iMessenger == NULL) {
		AppReturnValue = B_NO_MEMORY;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}
	
	for (int32 tIndex( 0 ); tIndex < 64; ++tIndex) {
		Connection[tIndex].Producer = media_node::null.node;
		Connection[tIndex].Consumer = media_node::null.node;
	}
	
	status_t tStatus;
	tStatus = ConstructConsumerNodes();
	if (tStatus < B_OK) {
		AppReturnValue = tStatus;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}

	iOpenFilePanel = new BFilePanel(
		B_OPEN_PANEL,
		iMessenger,
		NULL,
		0,
		true,
		new BMessage('OpnF'),
		NULL,
		false,
		true );
	if (iOpenFilePanel == NULL) {
		AppReturnValue = B_NO_MEMORY;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}

	BView* MainView(
		new BView( BWindow::Bounds(), NULL, B_FOLLOW_ALL, 0 ) );
	if (MainView == NULL)	{
		AppReturnValue = B_NO_MEMORY;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}
	MainView->SetViewColor( ui_color( B_PANEL_BACKGROUND_COLOR ) );
	BWindow::AddChild( MainView );
	
  iMenuBar = new BMenuBar(
  	BRect( 0.0f, 0.0f, 0.0f, 0.0f ),
  	"iMenuBar",
  	B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
  	B_ITEMS_IN_ROW,
  	true );
  if (iMenuBar == NULL) {
 		AppReturnValue = B_NO_MEMORY;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}

	BMenu * FileMenu = new BMenu(
		"File",
		B_ITEMS_IN_COLUMN );
	if (FileMenu == NULL) {
		AppReturnValue = B_NO_MEMORY;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}
	FileMenu->AddItem( new BMenuItem( "Open...", new BMessage('FOpn'), 'O', 0 ) );
	FileMenu->AddItem( new BMenuItem( "Quit", new BMessage('Quit'), 'Q', 0 ) );
	iMenuBar->AddItem( FileMenu );
	MainView->AddChild( iMenuBar );
	
	BButton * StartButton( new BButton(
		BRect( 10.0f, 10.0f, 40.0f, 30.0f ),
		"StartButton",
		"Start",
		new BMessage( 'Strt' ) ) );
	if (StartButton == NULL) {
		AppReturnValue = B_NO_MEMORY;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}
	MainView->AddChild( StartButton );
	StartButton->ResizeToPreferred();
	StartButton->MoveTo( 10.0f, iMenuBar->Frame().bottom + 10.0f );

	BButton * StopButton( new BButton(
		BRect( 10.0f, 10.0f, 40.0f, 30.0f ),
		"StopButton",
		"Stop",
		new BMessage( 'Stop' ) ) );
	if (StopButton == NULL) {
		AppReturnValue = B_NO_MEMORY;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}
	MainView->AddChild( StopButton );
	StopButton->ResizeToPreferred();
	StopButton->MoveTo( StartButton->Frame().left, StartButton->Frame().bottom + 10.0f );
	
	iSeekSlider = new BSlider( 
		BRect( MainView->Bounds().left + 10.0f, 10.0f, MainView->Bounds().right - 10.0f, 30.0f ),
		"SeekSlider",
		"Position",
		new BMessage('Seek'),
		0, 10000,
		B_TRIANGLE_THUMB,
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP );
	if (iSeekSlider == NULL) {
		AppReturnValue = B_NO_MEMORY;
		be_app->PostMessage( B_QUIT_REQUESTED );
		return;
	}
	MainView->AddChild( iSeekSlider );
	iSeekSlider->ResizeToPreferred();
	iSeekSlider->MoveTo( 10.0f, StopButton->Frame().bottom + 10.0f );
	
	iUpdateTimer = new BMessageRunner( *iMessenger, new BMessage('UpdT'), 500000, -1 );
}


clsMainWindow::~clsMainWindow() {
	DestructAllNodes();
	delete iUpdateTimer;
	delete iOpenFilePanel;
	delete iMessenger;
}


void clsMainWindow::MessageReceived( BMessage * Message ) {
	switch (Message->what) {
	case 'FOpn':
		iOpenFilePanel->Show();
		break;
	case 'OpnF':
		{
			entry_ref EntryRef;
			entry_ref FileRef;
			if (Message->FindRef("refs", 0, &EntryRef) == B_OK)
			{
				BEntry Entry(&EntryRef, true);
				BPath Path(&Entry);
				FileName = Path.Path();
			}
			if (ConstructProducerNode( FileName.String() ) < B_OK)
				Log("ERROR: Producer nodes could not be constructed.\n");
			else {
				if (Connect() < B_OK)
					Log("ERROR: Nodes could not be connected.\n");
			}
		}
		break;
	case 'Quit':
		AppReturnValue = B_OK;
		QuitRequested();
		break;
	case 'Strt':
		Start();
		break;
	case 'Stop':
		Stop();
		break;
	case 'UpdT':
		if ((iSeekSlider != NULL) && (iMediaPlayerNode != NULL)) {
			iSeekSlider->SetPosition(
				float( iMediaPlayerNode->Position() ) /
				float( iMediaPlayerNode->Duration() ) );
		}
		break;
	case 'Seek':
		if (MediaProducer != media_node::null) {
			int32 tPosition( 0 );
			Message->FindInt32( "be:value", &tPosition );
			iMediaRoster->SeekNode(
				MediaProducer,
				double( tPosition ) * 0.0001 * double( iMediaPlayerNode->Duration() ),
				0 );
		}
		break;
	default:
	  BWindow::MessageReceived( Message );
	  break;
	}
}

bool clsMainWindow::QuitRequested() {
	be_app->PostMessage( B_QUIT_REQUESTED );
	return BWindow::QuitRequested();
}


status_t clsMainWindow::ConstructConsumerNodes() {
	status_t tStatus;
	DestructAllNodes();

	tStatus = iMediaRoster->GetAudioMixer( &AudioConsumer );	
	if (tStatus < B_OK) {
		Log("ERROR: Could not get the audio consumer node.\n");
		DestructAllNodes();
		return tStatus;
	}
	
	tStatus = iMediaRoster->GetVideoOutput( &VideoConsumer );
	if (tStatus < B_OK) {
		Log("ERROR: Could not get the video consumer node.\n");
		DestructAllNodes();
		return tStatus;
	}

	tStatus = iMediaRoster->GetTimeSource( &TimeSource );
	if (tStatus < B_OK) {
		Log("ERROR: Could not get the preferred time source.\n");
		DestructAllNodes();
		return tStatus;
	}
	
	iTimeSourceNode = iMediaRoster->MakeTimeSourceFor( TimeSource );
	if (iTimeSourceNode == NULL) {
		Log("ERROR: Could not get a time source node for the preferred time source.\n");
		DestructAllNodes();
		return B_NO_MEMORY;
	}	

	return tStatus;
}


status_t clsMainWindow::ConstructProducerNode( const char * uFileName ) {
	status_t tStatus;
	DestructProducerNode();
	
	iMediaPlayerNode = new MediaPlayerNode( uFileName );
	if (iMediaPlayerNode == NULL)
		return B_NO_MEMORY;
	
	tStatus = iMediaRoster->RegisterNode( iMediaPlayerNode );
	if (tStatus < B_OK) {
		iMediaPlayerNode->Release();
		iMediaPlayerNode = NULL;
		DestructProducerNode();
		return tStatus;
	}
	
	tStatus = iMediaPlayerNode->InitCheck();	
	if (tStatus < B_OK) {
		Log("ERROR: MediaPlayerNode::InitCheck() Failed. (Bad File Name?)\n");
		DestructProducerNode();
		return tStatus;
	}
	
	tStatus = iMediaRoster->GetNodeFor( iMediaPlayerNode->ID(), &MediaProducer );
	if (tStatus < B_OK) {
		Log("ERROR: Could not get the video producer node.\n");
		DestructProducerNode();
		return tStatus;
	}
	
	return tStatus;	
}


status_t clsMainWindow::Connect() {
	status_t tStatus;
	Disconnect();

	int32 OutputListCount( 64 );
	media_output OutputList[OutputListCount];

	int32 tInputCount;
	media_input tInput;
	
	media_format tConnectionFormat;

	OutputListCount = 64;
	tStatus = iMediaRoster->GetFreeOutputsFor(
		MediaProducer,
		&OutputList[0],
		OutputListCount,
		&OutputListCount,
		B_MEDIA_RAW_AUDIO );
	if (tStatus < B_OK)	{
		Log("ERROR: Could not get audio outputs list.\n");
		OutputListCount = 0;
	}

	for (int32 tIndex( 0 ); tIndex < OutputListCount; ++tIndex) {
		tInputCount = 1;
		tStatus = iMediaRoster->GetFreeInputsFor(
			AudioConsumer,
			&tInput,
			tInputCount,
			&tInputCount,
			B_MEDIA_RAW_AUDIO );
		if ((tStatus >= B_OK) && (tInputCount >= 1)) {
			tConnectionFormat = OutputList[tIndex].format;
			tStatus = iMediaRoster->Connect(
				OutputList[tIndex].source,
				tInput.destination,
				&tConnectionFormat,
				&Connection[ConnectionCount].Output,
				&Connection[ConnectionCount].Input );
			if (tStatus >= B_OK) {
			  UsesAudio = true;
				Connection[ConnectionCount].Producer = MediaProducer.node;
				Connection[ConnectionCount].Consumer = AudioConsumer.node;
				ConnectionCount++;
			}
		} else {
			Log("ERROR: Could not get a free audio input.\n");
		}
	}	

	OutputListCount = 64 - OutputListCount;
	tStatus = iMediaRoster->GetFreeOutputsFor(
		MediaProducer,
		&OutputList[0],
		OutputListCount,
		&OutputListCount,
		B_MEDIA_RAW_VIDEO );
	if (tStatus < B_OK)	{
		Log("ERROR: Could not get video outputs list.\n");
		OutputListCount = 0;
	}

	for (int32 tIndex( 0 ); tIndex < OutputListCount; ++tIndex) {
		tInputCount = 1;
		tStatus = iMediaRoster->GetFreeInputsFor(
			VideoConsumer,
			&tInput,
			tInputCount,
			&tInputCount,
			B_MEDIA_RAW_VIDEO );
		if ((tStatus >= B_OK) && (tInputCount >= 1)) {
			tConnectionFormat = OutputList[tIndex].format;
			tStatus = iMediaRoster->Connect(
				OutputList[tIndex].source,
				tInput.destination,
				&tConnectionFormat,
				&Connection[ConnectionCount].Output,
				&Connection[ConnectionCount].Input );
			if (tStatus >= B_OK) {
			  UsesVideo = true;
				Connection[ConnectionCount].Producer = MediaProducer.node;
				Connection[ConnectionCount].Consumer = VideoConsumer.node;
				ConnectionCount++;
			}
		} else {
			Log("ERROR: Could not get a free audio input.\n");
		}
	}
	
	return tStatus;
}


status_t clsMainWindow::Start() {
	status_t tStatus( B_OK ), tStatus2( B_OK );

	if (TimeSource == media_node::null)
		return B_MEDIA_BAD_NODE;
		
	if (MediaProducer == media_node::null)
		return B_MEDIA_BAD_NODE;
	
	tStatus2 = iMediaRoster->SetTimeSourceFor( MediaProducer.node, TimeSource.node );
	if (tStatus2 < tStatus) tStatus = tStatus2;
	if (UsesAudio && (AudioConsumer != media_node::null)) {
		tStatus2 = iMediaRoster->SetTimeSourceFor( AudioConsumer.node, TimeSource.node );
		if (tStatus2 < tStatus) tStatus = tStatus2;
	}
	if (UsesVideo && (VideoConsumer != media_node::null)) {
		tStatus2 = iMediaRoster->SetTimeSourceFor( VideoConsumer.node, TimeSource.node );
		if (tStatus2 < tStatus) tStatus = tStatus2;
	}
	
	tStatus2 = iMediaRoster->PrerollNode( MediaProducer );
	if (tStatus2 < tStatus) tStatus = tStatus2;
	if (UsesAudio && (AudioConsumer != media_node::null)) {
		tStatus2 = iMediaRoster->PrerollNode( AudioConsumer );
		if (tStatus2 < tStatus) tStatus = tStatus2;
	}
	if (UsesVideo && (VideoConsumer != media_node::null)) {
		tStatus2 = iMediaRoster->PrerollNode( VideoConsumer );
		if (tStatus2 < tStatus) tStatus = tStatus2;
	}

	bigtime_t TotalLatency;
	tStatus2 = iMediaRoster->GetLatencyFor( MediaProducer, &TotalLatency );
	if (tStatus2 < tStatus) tStatus = tStatus2;

	bigtime_t StartTime;
	StartTime = TotalLatency + iTimeSourceNode->PerformanceTimeFor(
		BTimeSource::RealTime() + 10000 );

	tStatus2 = iMediaRoster->StartNode( MediaProducer, StartTime );
	if (tStatus2 < tStatus) tStatus = tStatus2;
	if (UsesAudio && (AudioConsumer != media_node::null)) {
		tStatus2 = iMediaRoster->StartNode( AudioConsumer, StartTime );
		if (tStatus2 < tStatus) tStatus = tStatus2;
	}
	if (UsesVideo && (VideoConsumer != media_node::null)) {
		tStatus2 = iMediaRoster->StartNode( VideoConsumer, StartTime );
		if (tStatus2 < tStatus) tStatus = tStatus2;
	}
	
	return tStatus;
}


status_t clsMainWindow::Stop() {
	status_t tStatus( B_OK ), tStatus2( B_OK );
	if (MediaProducer != media_node::null)
		tStatus2 = iMediaRoster->StopNode( MediaProducer, 0, true );
	if (tStatus2 < tStatus) tStatus = tStatus2;
//HACK: We don't want to stop the audio mixer, because it does Bad Things(tm)...
//	if (UsesAudio && (AudioConsumer != media_node::null))
//		tStatus2 = iMediaRoster->StopNode( AudioConsumer, 0, true );
	if (tStatus2 < tStatus) tStatus = tStatus2;
	if (UsesVideo && (VideoConsumer != media_node::null))
		tStatus2 = iMediaRoster->StopNode( VideoConsumer, 0, true );
	if (tStatus2 < tStatus) tStatus = tStatus2;
	return tStatus;
}


status_t clsMainWindow::Disconnect() {
	status_t tStatus, tStatus2( B_OK );
	tStatus = Stop();

	UsesAudio = false;
	UsesVideo = false;
	
	for (int32 tIndex( 0 ); tIndex < ConnectionCount; ++tIndex) {
		tStatus2 = iMediaRoster->Disconnect(
				Connection[tIndex].Producer,
				Connection[tIndex].Output.source,
				Connection[tIndex].Consumer,
				Connection[tIndex].Input.destination
			);
		if (tStatus2 < tStatus) tStatus = tStatus2;
		Connection[tIndex].Producer = media_node::null.node;
		Connection[tIndex].Consumer = media_node::null.node;
	}
	ConnectionCount = 0;
	
	return tStatus;
}


status_t clsMainWindow::DestructAllNodes() {
	status_t tStatus, tStatus2( B_OK );
	tStatus = Disconnect();
	
	tStatus2 = DestructProducerNode();
	if (tStatus2 < tStatus) tStatus = tStatus2;

	if (iTimeSourceNode != NULL) {
	  //HACK: We don't do this here, because the iMediaRoster->ReleaseNode( TimeSource )
	  //call down below does the same thing...
		//iTimeSourceNode->Release();
		iTimeSourceNode = NULL;
	}
	if (AudioConsumer != media_node::null) {
		tStatus = iMediaRoster->ReleaseNode( AudioConsumer );
		if (tStatus2 < tStatus) tStatus = tStatus2;
		AudioConsumer = media_node::null;
	}
	if (VideoConsumer != media_node::null) {
		tStatus = iMediaRoster->ReleaseNode( VideoConsumer );
		if (tStatus2 < tStatus) tStatus = tStatus2;
		VideoConsumer = media_node::null;
	}
	if (TimeSource != media_node::null) {
		tStatus = iMediaRoster->ReleaseNode( TimeSource );
		if (tStatus2 < tStatus) tStatus = tStatus2;
		TimeSource = media_node::null;
	}

	return tStatus;
}


status_t clsMainWindow::DestructProducerNode() {
	status_t tStatus, tStatus2( B_OK );
	tStatus = Disconnect();
	
	if (iMediaPlayerNode != NULL) {
		tStatus2 = iMediaRoster->UnregisterNode( iMediaPlayerNode );
		if (tStatus2 < tStatus) tStatus = tStatus2;
		iMediaPlayerNode->Release();
		iMediaPlayerNode = NULL;
	}

	if (MediaProducer != media_node::null) {	
		tStatus = iMediaRoster->ReleaseNode( MediaProducer );
		if (tStatus2 < tStatus) tStatus = tStatus2;
		MediaProducer = media_node::null;
	}

	return tStatus;
}
