//Title: Graphics Device Class
//Platform: BeOS 5
//Version: 2001.02.03
//Description:
//	A class for working with physical graphics devices (like VGA displays)
//  Implements a high-level interface to the BeOS graphics drivers and accelerants

//Copyright (C) 2002 Kevin H. Patterson

#include "GraphicsDevice.h"

#include <iostream.h>
	
GraphicsDevice::GraphicsDevice(const char * DeviceName)
:
	iDeviceName(DeviceName),
	iInitCheck(false),
	hDevice(-1),
	AccelerantImage(-1),
	GetAccelerantFunction(NULL),
	AccelerantDeviceInfo(new accelerant_device_info)
{
	//open the driver...
	hDevice = open(DeviceName, B_READ_WRITE);
	if (hDevice >= 0) {
		//device driver has been opened sucessfully...

		//load the accelerant image...
		AccelerantImage = LoadAccelerantImage();
		if (AccelerantImage >= 0) {
			//accelerant image has been loaded successfully...

			iInitCheck = true;

			get_accelerant_device_info GetAccelerantDeviceInfo(NULL);
			GetAccelerantDeviceInfo = (get_accelerant_device_info)GetAccelerantFunction(B_GET_ACCELERANT_DEVICE_INFO, NULL);
			
			if (GetAccelerantDeviceInfo != NULL) {
				//HACK: I have to call this function twice for some STRANGE reason!
				GetAccelerantDeviceInfo(AccelerantDeviceInfo);
				GetAccelerantDeviceInfo(AccelerantDeviceInfo);
			}

		}
		else {
			close(hDevice);
			hDevice = -1;	
		}
	}
}

GraphicsDevice::~GraphicsDevice() {
	delete AccelerantDeviceInfo;

	//unload the accelerant image...
	if (AccelerantImage >= 0) {
		//un-initialize the accelerant...
		uninit_accelerant UnInitAccelerant(NULL);
		UnInitAccelerant = (uninit_accelerant)GetAccelerantFunction(B_UNINIT_ACCELERANT, NULL);
		if (UnInitAccelerant != NULL)
			UnInitAccelerant();
		
		//unload the accelerant image...
		unload_add_on(AccelerantImage);
	}
	
	//close the driver...
	if (hDevice >= 0)
		close(hDevice);
}

bool GraphicsDevice::InitCheck() {
	return iInitCheck;
}

const char * GraphicsDevice::DeviceName() {
	return iDeviceName;
}

const char * GraphicsDevice::Name() {
	return AccelerantDeviceInfo->name;
}

const char * GraphicsDevice::Chipset() {
	return AccelerantDeviceInfo->chipset;
}

const char * GraphicsDevice::SerialNumber() {
	return AccelerantDeviceInfo->serial_no;
}

uint32 GraphicsDevice::Memory() {
	return AccelerantDeviceInfo->memory;
}

uint32 GraphicsDevice::DACSpeed() {
	return AccelerantDeviceInfo->dac_speed;
}

BBufferGroup * GraphicsDevice::CreateBufferGroup() {

	get_frame_buffer_config gfbc;
	gfbc = (get_frame_buffer_config)GetAccelerantFunction(B_GET_FRAME_BUFFER_CONFIG, NULL);
	frame_buffer_config a_frame_buffer;
	if (gfbc == NULL)
		return NULL;
	if (gfbc(&a_frame_buffer) != B_OK)
		return NULL;

	area_id MyFrameBuffer;
	bool WeCreatedIt(false);

	MyFrameBuffer = area_for(a_frame_buffer.frame_buffer);
	if (MyFrameBuffer < 0) {
		MyFrameBuffer = create_area("MyFrameBuffer", &a_frame_buffer.frame_buffer, B_EXACT_ADDRESS, 307200 * 4, B_CONTIGUOUS, B_READ_AREA | B_WRITE_AREA);
		if (MyFrameBuffer >=0) WeCreatedIt = true;
	}

	if (MyFrameBuffer >=0) {
		printf("We got it.\n");
		area_info ai;
		get_area_info(MyFrameBuffer, &ai);
		cout << "Area ID: " << ai.area << endl;
		cout << "Name: " << ai.name << endl;
		cout << "Size (K): " << (ai.size / 1024) << endl;
		cout << "Locked Size (K): " << (ai.ram_size / 1024) << endl;
		cout << "Team Address: " << ai.address << endl;
	}
	else {
		printf(strerror(MyFrameBuffer));
		printf("\n");
	}

	if (WeCreatedIt) printf("We actually created it.\n");

	if (WeCreatedIt) delete_area(MyFrameBuffer);
	
	//BBufferGroup *my_group = new BBufferGroup; 
/*	BBitmap *my_bitmap = new BBitmap(BRect(0,0,639,479), B_BITMAP_IS_AREA, B_RGB32, 640*4); 
	
	if (!my_bitmap->IsValid()) { 
		return B_ERROR; 
	} 
	area_info bm_info; 
	
	if (get_area_info(my_bitmap->Area(), &bm_info) != B_OK) { 
		return B_ERROR; 
	} 
	
	buffer_clone_info bc_info; 
	bc_info.area = bm_info.area; 
	bc_info.offset = ((char *) my_bitmap->Bits())-((char *) bm_info.address); 
	bc_info.size = my_bitmap->BitsLength(); 
	
	BBuffer *out_buffer = NULL; 
	status_t err = my_group->AddBuffer(bc_info, &out_buffer); 
*/	
	/* out_buffer is now set to use the BBitmap's memory */
	//return my_group;
	return NULL;
}

void GraphicsDevice::DeleteBufferGroup(BBufferGroup * uBufferGroup) {
}

image_id GraphicsDevice::LoadAccelerantImage() {
	status_t result;
	image_id image = -1;
	uint i;
	char
		signature[1024],
		path[PATH_MAX];
	struct stat st;
	const static directory_which vols[] = {
		B_USER_ADDONS_DIRECTORY,
		B_COMMON_ADDONS_DIRECTORY,
		B_BEOS_ADDONS_DIRECTORY
	};

	//get the accelerant signature from the driver...
	result = ioctl(hDevice, B_GET_ACCELERANT_SIGNATURE, &signature, sizeof(signature));
	if (result != B_OK) goto done;

	for (i=0; i < sizeof (vols) / sizeof (vols[0]); i++) {
		//	compute directory path to common or beos addon directory on
		//	floppy or boot volume

  	if (find_directory (vols[i], -1, false, path, PATH_MAX) != B_OK) continue;

		strcat (path, "/accelerants/");
		strcat (path, signature);

		// don't try to load non-existant files
		if (stat(path, &st) != 0) continue;

		// load the image
		image = load_add_on(path);
		if (image >= 0) {

			// get entrypoint from accelerant
			result = get_image_symbol (
				image,
				B_ACCELERANT_ENTRY_POINT,
#if defined(__INTEL__)
				B_SYMBOL_TYPE_ANY,
#else
				B_SYMBOL_TYPE_TEXT,
#endif
				(void **)&GetAccelerantFunction
			);

			if (result == B_OK) {
				init_accelerant ia;
				ia = (init_accelerant)GetAccelerantFunction(B_INIT_ACCELERANT, NULL);
				if (ia && ((result = ia(hDevice)) == B_OK)) {
					// we have a winner!
					break;
				}
			}

			// unload the accelerant, as we must be able to init!
			unload_add_on(image); 
		}

		// mark failure to load image
		image = -1;
	}

done:
	return image;
}

bool GraphicsDevice::SetTVOutMode(const int Mode, const int OutputMode = 3, const bool FlickerFilter = true, const bool DualDisplay = true) {

	int monstat = 0, 												//'nothing' connected
		bus0wired = 0, bus1wired = 0,					//no wired busses found
		bus0sane = 0, bus1sane = 0,						//no sane buses
		fd,																		//driver pointer
		Version = 0xff,												//TVout chip silicon-version
		Type = 0xff,													//TVout chiptype
		stat = B_OK;

	unsigned char OutMethod = OutputMode,		//TV output signals are Y/C and CVBS...
		ffilter = (FlickerFilter & 1),				//flicker filter is default ON...
		input = 0,
		dualclone = (DualDisplay & 1);				//Singlehead output is default selected...

	//create device data struct...
	riva_data_t *ddata = new riva_data_t;

	display_mode org_mode, new_mode;

	fd = open ("/dev/video/riva_iic/1", O_RDWR);
	
	//check if driver could be reached...
	if (fd < 0)
		return (B_ERROR);

	//setup 'magic value' to unlock driver: 
	//(to keep compatibility & prevent crashes with older versions of driver and app...)
	ddata->drv_version = 0xff;	
	//notify driver of our version:
	//(to prevent old apps from accessing new drivers...)
	ddata->app_version = (uint16) VERSION;
	//check version of driver...
	stat = (ioctl(fd, rivaVersion, ddata));

	if (ddata->drv_version == 0xff)
	{
		printf("riva_iic driver V0.00\n");
		printf("\nIncompatible driver version (V0.22 or higher needed), exiting...\n");
		close (fd);
		return (B_OK);
	}
	//this works without driver supporting riva_data_t struct:
	//well, it did before I changed the first var-type.. :-(
	printf("riva_iic driver V%.2f\n",ddata->drv_version/100.0); 
	if (ddata->drv_version < 22)
	{
		printf("\nIncompatible driver version (V0.22 or higher needed), exiting...\n");
		close (fd);
		return (B_OK);
	}

	//(now we know the driver supports the riva_data_t struct...)
	if (stat) //if ioctl returned 'NACK'..
	{
		printf("\nBeTVOut application too old for use with this driver, exiting...\n"
				"(Version %.3f or higher needed.)\n",ddata->app_version/1000.0);
		close (fd);
		return (B_OK);
	}

	//unlock RIVA extended VGA registers...
	ioctl(fd, rivaUnlock);

	//check for one or two available (wired) IIC busses.
	//On RIVA TNT and later, two could be there...
	int RIVAexists = 0;
	printf("RIVA IICbus 0: ");
	ddata->iic.busnr = 0;
	if (!ioctl(fd, rivaChkBus, ddata))
	{
		printf("detected.\n");
		RIVAexists = 1;
		bus0wired = 1;
	}
	else 
	{
		printf("not detected.\n");
	}
	printf("RIVA IICbus 1: ");
	ddata->iic.busnr = 1;
	if (!ioctl(fd, rivaChkBus, ddata))
	{
		printf("detected.\n");
		RIVAexists = 1;
		bus1wired = 1;
	}
	else
	{
		printf("not detected.\n");
	}
	if (!RIVAexists)
	{
		printf("No RIVA IIC bus found, exiting...\n");
		close (fd);
		return(B_OK);
	}

	//find at least one sane IIC bus... (test this by sending an unused adress...)
	if (!FindSaneBus (fd, bus0wired, bus1wired, &bus0sane, &bus1sane, ddata))
	{
		printf("Unexpected results, exiting...\n");
		close (fd);
		return (B_OK);
	}

	//now find a TV-out chip...
	int	chfound = 0;
	byte chbus = 0, chaddr = 0, chversion;

	//check for BT or CX chip:
	int btfound = FindBT (fd, ddata);
	if (btfound)
	{
		//read chiptype and version number:
		if ((input = ReadType (fd, &Version, &Type, ddata)))
		{
			//something went wrong:
			if (input == 1)
			{
				//IIC error encountered:
				return (B_OK);
			}
			else
			{
				//The chip apparantly is not a TVout chip:
				btfound = 0;
			}	
		} 
		else
		{
			//no errors found:
			printf("ChipID = %d, chip version = %d: Chiptype is ",Type,Version);
			switch (Type)
			{
				case 0: printf("bt868.\n");
						break;
				case 1: printf("bt869.\n");
						break;
				case 2: printf("cx25870.\n");
						break;
				case 3: printf("cx25871.\n");
						break;
				default:printf("unknown, exiting...\n");
						close (fd);
						return(0);
			}
		}
	}		

	//check for CH chip:
	if (!btfound)
	{
		if (bus0sane)
		{
			ddata->iic.busnr = 0;
			chbus = 0;
			chfound = ch_detect(fd, chbus, &chaddr, &chversion, ddata);
		}
		if (bus1sane && !chfound)
		{
			ddata->iic.busnr = 1;
			chbus = 1;
			chfound = ch_detect(fd, chbus, &chaddr, &chversion, ddata);
		}

		if (chfound)
		{
			printf("\nFound one on bus %d, adress $%x;\n",chbus,chaddr);
			printf("Chiptype is ch%s.\n",ch_model_name(chversion));
		}
		else //no TVout chip at all...
		{
			printf("\nNo supported TVout chip found, exiting...\n");
			close (fd);
			return(B_OK);
		}
	}

	get_display_mode gdm;
	gdm = (get_display_mode)GetAccelerantFunction(B_SET_DISPLAY_MODE, NULL);
	gdm(&org_mode);		//'save' current mode...
	new_mode = org_mode;

//KHP

	ddata->devID = 0;
	//get current RIVA display mode:
	ioctl(fd, rivaGetMode, ddata);

	printf("(RIVA h_display: %d, ",ddata->displ_mode.displ_timing.h_display);
	printf("v_display: %d)\n",ddata->displ_mode.displ_timing.v_display);

	//RIVA programming should be universal for every chipversion, but
	//the 'horizontal sync phase' setting in the TV-out chip SHOULD be done 
	//seperately for different RIVA chips and videomodes though...
	//(for both CH and BT chips).
	//WARNING: This applies for both VESA and 'driver' (BeScreen) mode!
	printf("Vendor ID = $%x, Device ID = $%x;\n",
									ddata->dev.vendor_id, ddata->dev.device_id);
	printf("Nvidia Corporation");
	switch (ddata->dev.device_id)
	{
		case 0x0020:	printf(" RIVA TNT1 detected.\n");
						ddata->type = 1;
						break;
		case 0x0028:	printf(" RIVA TNT2 detected.\n");
						ddata->type = 2;
						break;
		case 0x002d:	printf(" RIVA TNT2-M64 detected.\n");
						ddata->type = 3;
						break;
		case 0x0110:	printf(" RIVA GeForce2 MX(200/400) detected.\n");
						ddata->type = 1;
						break;
		default:		printf(": Untested card detected!\n");
						printf("\nPlease test BeTVOut to find the most optimal timing\n"
								"option for use with your card and then inform us of\n"
								"the results so we can automate setup for you in the\n"
								"next BeTVOut version...\n");
						printf("\nSelect chip specific timing:\n");
						printf("1. TNT1 or GeForce2 compatible;\n");
						printf("2. TNT2 compatible;\n");
						printf("3. TNT2-M64 compatible.\n");
						printf("Option: ");
						fflush(stdin);
						do scanf ("%c",&ddata->type);
						while ((ddata->type < '1') | (ddata->type > '3'));
						ddata->type -= 48; 
	}

	//What video signals should be outputted?
	//(only valid for BT, CH is unable to cross-switch outputs..)
	if (!btfound) OutMethod = 3; //CHfound: detect actual connections anyway for 'statusinfo'..

	input = Mode;
	switch (input)
	{
		case 1:
		case 4:
		//HACK: What is this for? Delete it...
		/*	if (VGAonlyMode (fd, Type, btfound, chbus, chaddr, screenptr,
																 org_mode,ddata))
				{
					close (fd);
					return (B_OK);
				} */
			break;
		case 2:
		case 3:
		case 5:
		case 6:
		case 9: 
		case 10:
		case 11:
		case 12://ask for 'singlehead' or 'dualhead clone' mode:
				dualclone = (DualDisplay & 1);
				//ask for use of 'flicker filter':
				ffilter = (FlickerFilter & 1);
				break;
		case 8: //exit to VGA only.
		//HACK: Not this time...
		/*		printf("\nExiting: ");
				//select 'VGA only' mode:
				if (VGAonlyMode (fd, Type, btfound, chbus, chaddr, screenptr,
																 org_mode, ddata))
				{
					close (fd);
					return (B_OK);
				} */
				close (fd);
				return(B_OK);
				break;
		default:;//nothing to be done here...
	}	

	//OK, let's do it...
	if (btfound)
	{
		//fixme: reset BT should be here...
		//(but: beware of the 'locked SDA' syndrome then!)
		if (KillClock_Blackout_BT (fd, Type, ddata)) return (B_OK);

		//if TV output signals is set to 'autodetect' then do it:
		if (OutMethod == 3)
		{
			monstat = ReadBT_ConnectStat (fd, ddata);
			if (monstat == -1) return (B_OK);
		}

		//'slowdown RIVA clock' if TVout is requested:
		//makes sure TVout chip is not being overclocked! (VESA640/800 is safe..)
		//NOTE: it looks like this is unnesessary after all..
		switch (input)
		{
			case 2:
			case 5:
			case 11:
				nonVESAsetVESA640(new_mode);
				break;
			case 3:
			case 6:
			case 9:
			case 10:
			case 12:
				nonVESAsetVESA800(new_mode);
				break;
			default:;//nothing to be done here...
		}

		//Manual config is non-dependent of the state of the PAL hardware input pin;
		//Also SDA lockups occur when setting EN_XCLK after autoconfig!
		//Make sure PAL_MD=0 for NTSC and PAL_MD = 1 for PAL...
		switch (input)
		{
			case 1: if (InitBT_NTSC640 (fd, ddata)) return (B_OK);
					break;
			case 2: if (InitBT_NTSC640 (fd, ddata)) return (B_OK);
					break;
			case 3: if (InitBT_NTSC800 (fd, ddata)) return (B_OK);
					break;
			case 4: if (InitBT_PAL800 (fd, ddata)) return (B_OK);
					break;
			case 5: if (InitBT_PAL640 (fd, ddata)) return (B_OK);
					break;
			case 6: if (InitBT_PAL800 (fd, ddata)) return (B_OK);
					break;
			case 9:	if (InitBT_NTSC720 (fd, ddata)) return (B_OK);
					break;
			case 10:if (InitBT_PAL720 (fd, ddata)) return (B_OK);
					break;
			case 11:if (InitBT_NTSC640_OS (fd, ddata)) return (B_OK);
					break;
			case 12:if (InitBT_PAL800_OS (fd, ddata)) return (B_OK);
					break;
			default:;//nothing to be done here...
		}	

		//modify BT Hphase signal to center TV image...
		if (SetBT_Hphase (fd, input, ddata)) return (B_OK);

		//setup BT for TVout (filters, DAC_muxes...)
		if (SetupBTforTVout (fd, OutMethod, monstat, ffilter, ddata)) return (B_OK);
    	printf("\nOption sent...");
	}
	else //CH found...
	{
		//reset CH or use another way to make sure all registers get set the same
		//way non-dependent of how they were set before BeTVOut was started...

		//if TV output signals is set to 'autodetect' then do it:
		//(make sure this does only gets called if CH input clock gets set below
		// its speed limits afterwards, or CH might get destroyed...)
		if ((OutMethod == 3) & (!(input == 1)) & (!(input == 4)))
		{
			monstat = ch_detect_connection(fd, chbus, chaddr, ddata);
			printf("\n(TV output monitor status = %d)\n",monstat);
		}
		//monstat value is:
		//0: Y/C and CVBS is connected
		//1: only Y/C is connected
		//6: only CVBS is connected
		//7: nothing is connected.
		//(tested with ASUS V7100 GeForce2 MX200 AGP 32Mb.)

		//'slowdown RIVA clock' if TVout is requested:
		//makes sure TVout chip is not being overclocked! (VESA640/800 is safe..)
		switch (input)
		{
			case 2:
			case 5:
			case 11:
				nonVESAsetVESA640(new_mode);
				break;
			case 3:
			case 6:
			case 9:
			case 10:
			case 12:
				nonVESAsetVESA800(new_mode);
				break;
			default:
				break;	//nothing to be done here...
		}

		//For TNT2 chips horizontal position should be shifted right by 8 pixels.
		//(all values tested on TNT2 and GeForce2 cards.)
		unsigned char hoffset = 2;		//2 x 4 pixels: TNT2 or TNT2 M64 
		if (ddata->type == 1) hoffset = 0;	//TNT1 or GeForce2

		//Because DVD modes with Chrontel are critical a special correction
		//needs to be done for these modes if RIVA is a TNT1 or GeForce2...
		//(It looks like the non-M64/M64 patch is needed, but this doesn't seem to
		// work here. The non-M64/M64 patch correction is normally visible only on
		// non-overscanning modes, so it only needs to be done there if these
		// modes will be set using BeScreen which currently isn't the case.
		// The currently used direct register programming method *is* using the
		// non-M64/M64 patch.
		// NOTE that the normally used 'hoffset' settings are set for centered TV
		// output, including for the Chrontel DVD modes. Only there we get
		// viewport problems if we don't apply the patch below:)
		if (((input == 9) | (input == 10)) & (ddata->type == 1)) hoffset = 2;

		switch (input)
		{
			case 1: break;//testimage not possible on CH
			case 2: //NTSC640 is mode 17, this is compatible to BT and RIVA timing.
					ch_rud_set_mode	(fd, chbus, chaddr,17,(0x30 + hoffset),
													0x02,0x01, ffilter, ddata);
													//mode, horpos, verpos, oflow
					break;
			case 3: //NTSC800 resembles mode 24.
					ch_rud_set_mode
						(fd, chbus, chaddr,24,(0x3e + hoffset),
													0xff,0x00, ffilter, ddata);
													//mode, horpos, verpos, oflow
					break;
			case 4: break;//testimage not possible on CH
			case 5: //PAL640 resembles mode 19.
					ch_rud_set_mode
						(fd, chbus, chaddr,19,(0x38 + hoffset),
													0x00,0x00, ffilter, ddata);
													//mode, horpos, verpos, oflow
					break;
			case 6: //PAL800 is mode 20, this is compatible to BT and RIVA timing.
					ch_rud_set_mode
						(fd, chbus, chaddr,20,(0x40 + hoffset),
													0x00,0x00, ffilter, ddata);
													//mode, horpos, verpos, oflow
					break;
			case 9:	//CH DVD MODES NOTICE:
					//The official CH DVD modes are interlaced input only, while
					//the RIVA cannot output interlaced data. This is why these
					//CH modes (25 and 26) cannot be used..  :-(
					//SO: if you want 'perfect' TV output in non-standard modes:
					//get a card with Brooktree BT868/869 chip mounted on it...

					//NTSC720 'resembles' mode 16.
					//(tested on TNT2-M64 with CH7006 only: has to be tested more..)
					//WARNING: do not change Hoffset: CH viewport is critical,
					//because we 'display' 720 pixels on a 640 pixel wide viewport.
					//(current values are (and have to be) exactly 'balanced'.)
					ch_rud_set_mode
						(fd, chbus, chaddr,16,(0x16 + hoffset),
													0x00,0x01, ffilter, ddata);
													//mode, horpos, verpos, oflow
					break;
			case 10://PAL720 resembles mode 19.
					//(tested on TNT2-M64 with CH7006 only: has to be tested more..)
					//WARNING: be carefull if Hoffset in changed: CH H_overscan is
					//critical, because we display 720 pixels in a 800 pixel
					//viewport.
					//(current values are (and have to be) exactly 'balanced'.)
					ch_rud_set_mode
						(fd, chbus, chaddr,19,(0x14 + hoffset),
													0x00,0x00, ffilter, ddata);
													//mode, horpos, verpos, oflow
					break;
			case 11://NTSC640_OS is mode 16, this is compatible to BT and RIVA timing.
					ch_rud_set_mode
						(fd, chbus, chaddr,16,(0x22 + hoffset),
													0x00,0x01, ffilter, ddata);
													//mode, horpos, verpos, oflow
					break;
			case 12://PAL768 resembles mode 19.
					ch_rud_set_mode
						(fd, chbus, chaddr,19,(0x1a + hoffset),
													0x00,0x00, ffilter, ddata);
													//mode, horpos, verpos, oflow
					break;
			default:;//nothing to be done here...
		//BTW: I checked the videosignals on a scope:
		//all used modes are all perfectly OK!

		//BTW 2: The CH700X seem to be much more forgiving than the BT chips!
		//If you change just a little in the wrong way on the BT your TV and VGA
		//image dissappear, while the CH does not suffer from these things!
		//Also the CH pixelclock may be as high as 50Mhz, while BT max is only 40Mhz..
		//The downside: The CH only supports the predefined modes, while the BT
		//is (almost) fully programmable...

		//BTW3: Above picture modes and position are tested on a
		//ASUS AGP 32Mb V7100 GeForce2 MX200 with CH7007 TVout chip,
		//and on a TNT2 card.
		}	

		//modify Hphase signal to center TV image...
		//already done in 'ch_rud_set_mode' above.

		//setup CH for TVout (filters...)
	    printf("\nOption sent...");
	}

	switch (input)//both BT and CH get here..
	{
		case 1: if (btfound)
				{
					if (DoTestImage (fd, ddata)) return (B_OK);
				   	printf(" and NTSC testimage activated.\n");
				}
				else //chfound
				{
				   	printf("\n\nSelected function not possible on this system.\n");
				}
				break;
		case 2: //reprogram RIVA for NTSC 640x480 @ 60Hz mode
				ioctl(fd, rivaNtsc640, ddata);
				printf(" and NTSC 640x480 @ 60Hz mode activated.\n");
				break;
		case 3: //reprogram RIVA for NTSC 800x600 @ 60Hz mode
				ioctl(fd, rivaNtsc800, ddata);
				printf(" and NTSC 800x600 @ 60Hz mode activated.\n");
				break;
		case 4: if (btfound)
				{
					if (DoTestImage (fd, ddata)) return (B_OK);
			   		printf(" and PAL testimage activated.\n"); 
				}
				else //chfound
				{
				   	printf("\n\nSelected function not possible on this system.\n");
				}
				break;
		case 5: //reprogram RIVA for PAL 640x480 @ 50Hz mode
				ioctl(fd, rivaPal640, ddata);
				printf(" and PAL 640x480 @ 50Hz mode activated.\n");
				break;
		case 6: //reprogram RIVA for PAL 800x600 @ 50Hz mode
				ioctl(fd, rivaPal800, ddata);
				printf(" and PAL 800x600 @ 50Hz mode activated.\n");
				break;
		case 9:	if (btfound)
				{
					//reprogram RIVA for NTSC 720x480 @ 60Hz mode:
					//(tested on TNT2 with BT869)
					new_mode.virtual_width = 720;			//H_ACTIVE
					new_mode.timing.h_display = 720;
					new_mode.virtual_height = 480;			//V_ACTIVEI
					new_mode.timing.v_display = 480;

					new_mode.timing.h_sync_start = 744;		//do not change!
					new_mode.timing.h_sync_end = 744+144;	//delta is H_sync_pulse
					new_mode.timing.h_total = 888;			//BT H_TOTAL
					new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
					new_mode.timing.v_sync_end = 490+26;	//delta is V_sync_pulse
					new_mode.timing.v_total = 525;			//CH V_TOTAL
					new_mode.timing.pixel_clock = ((888 * 525 * 60) / 1000); //refresh
				}
				else //CH found
				{
					//reprogram RIVA for NTSC 720x480 @ 60Hz mode:
					//(tested on TNT2-M64 with CH7006)
					new_mode.virtual_width = 720;			//H_ACTIVE
					new_mode.timing.h_display = 720;
					new_mode.virtual_height = 480;			//V_ACTIVEI
					new_mode.timing.v_display = 480;

					new_mode.timing.h_sync_start = 744;		//do not change! (CH viewport is critical...)
					new_mode.timing.h_sync_end = 744+20;	//delta is H_sync_pulse
					new_mode.timing.h_total = 784;			//CH H_TOTAL
					new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
					new_mode.timing.v_sync_end = 490+25;	//delta is V_sync_pulse
					new_mode.timing.v_total = 525;			//CH V_TOTAL
					new_mode.timing.pixel_clock = ((784 * 525 * 60) / 1000); //refresh
				}
				ScreenSetMode(&new_mode, false);
				//switch RIVA to slave-mode for TVout chip:
				ioctl(fd, rivaEnableTVOut);
				printf(" and NTSC 720x480 @ 60Hz overscanning mode activated.\n");
				break;
		case 10://reprogram RIVA for PAL 720x576 @ 50Hz mode:(a bit modified for CH)
				new_mode.virtual_width = 720;			//BT H_ACTIVE
				new_mode.timing.h_display = 720;
				new_mode.virtual_height = 576;			//BT  V_ACTIVEI
				new_mode.timing.v_display = 576;
				new_mode.timing.h_sync_start = 744;		//set for centered sync pulse
				new_mode.timing.h_sync_end = 744+140;	//delta is BT H_BLANKI
				new_mode.timing.h_total = 888;			//BT H_CLKI
				new_mode.timing.v_sync_start = 579;		//set for centered sync pulse
				new_mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
				new_mode.timing.v_total = 625;			//BT V_LINESI
				new_mode.timing.pixel_clock = ((888 * 625 * 50) / 1000); //refresh
				ScreenSetMode(&new_mode, false);
				//switch RIVA to slave-mode for TVout chip:
				ioctl(fd, rivaEnableTVOut);
				printf(" and PAL 720x576 @ 50Hz overscanning mode activated.\n");
				break;
		case 11://reprogram RIVA for NTSC 640x480 (OverScanning) @ 60Hz mode:
				new_mode.virtual_width = 640;			//BT H_ACTIVE
				new_mode.timing.h_display = 640;
				new_mode.virtual_height = 480;			//BT  V_ACTIVEI
				new_mode.timing.v_display = 480;
				new_mode.timing.h_sync_start = 744;		//set for CH/BT compatible TV output
				new_mode.timing.h_sync_end = 744+20;	//delta is BT H_BLANKI
				new_mode.timing.h_total = 784;			//BT H_CLKI
				new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
				new_mode.timing.v_sync_end = 490+25;	//delta is BT V_BLANKI
				new_mode.timing.v_total = 525;			//BT V_LINESI
				new_mode.timing.pixel_clock = ((784 * 525 * 60) / 1000); //refresh
				ScreenSetMode(&new_mode, false);
				//switch RIVA to slave-mode for TVout chip:
				ioctl(fd, rivaEnableTVOut);
				printf(" and NTSC 640x480 @ 60Hz overscanning mode activated.\n");
				break;
		case 12://reprogram RIVA for PAL 768x576 @ 50Hz mode:
				new_mode.virtual_width = 768;			//H_ACTIVE
				new_mode.timing.h_display = 768;
				new_mode.virtual_height = 576;			//V_ACTIVEI
				new_mode.timing.v_display = 576;
				new_mode.timing.h_sync_start = 848;		//set for centered TV output
				new_mode.timing.h_sync_end = 848+20;	//delta is BT H_BLANKI
				new_mode.timing.h_total = 944;			//BT H_CLKI
				new_mode.timing.v_sync_start = 579;		//set for centered sync pulse
				new_mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
				new_mode.timing.v_total = 625;			//BT V_LINESI
				new_mode.timing.pixel_clock = ((944 * 625 * 50) / 1000); //refresh
				ScreenSetMode(&new_mode, false);
				//switch RIVA to slave-mode for TVout chip:
				ioctl(fd, rivaEnableTVOut);
				printf(" and PAL 768x576 @ 50Hz overscanning mode activated.\n");
				break;
		default:;//nothing to be done here...
	}	

	switch (input)
	{
		case 2:
		case 3:
		case 5:
		case 6:
		case 9:
		case 10:
		case 11:
		case 12://select 'singlehead' or 'dualhead clone' mode:
				if (dualclone)
				{
					//turn on VGA:
					ioctl(fd, rivaVgaOn);
				}
				else
				{
					//turn off VGA:
					ioctl(fd, rivaVgaOff);
				}
				break;
		case 7: //select 'VGA only' mode:
			//HACK: YANH (Yet Another Hack)
			//	if (VGAonlyMode (fd, Type, btfound, chbus, chaddr, screenptr,
			//													 org_mode, ddata))
			//	{
					close (fd);
					return (B_OK);
			//	}
				break;
		default:;//nothing to be done here...
	}

}//end main.

int GraphicsDevice::nonVESAsetVESA640(display_mode new_mode)
{
	//set VESA640: (hex values are exact copies of direct register programmings)
	new_mode.virtual_width = 640;
	new_mode.virtual_height = 480;
	new_mode.timing.h_display = 640;
	new_mode.timing.v_display = 480;
	
	new_mode.timing.h_sync_start = (0x53 * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1f + (0x53 && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x5F * 8);//byte oriented! (val. *must* be set!)
	new_mode.timing.v_sync_start = 0x1ea;
	new_mode.timing.v_sync_end = ((0x1ea && 0xff0) + 0x00c);
	new_mode.timing.v_total = 0x20b;
	new_mode.timing.pixel_clock = ((((0x5F * 8) * 0x20b)* 60)/1000);// in kHz!

	return ScreenSetMode(&new_mode, false);
}//end nonVESAsetVESA640.

int GraphicsDevice::nonVESAsetVESA800(display_mode new_mode)
{
	//set VESA800: (hex values are exact copies of direct register programmings)
	new_mode.virtual_width = 800;
	new_mode.virtual_height = 600;
	new_mode.timing.h_display = 800;
	new_mode.timing.v_display = 600;
	
	new_mode.timing.h_sync_start = (0x6a * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1a + (0x6a && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x7f * 8);//byte oriented! (this value *must* be set!)
	new_mode.timing.v_sync_start = 0x259;
	new_mode.timing.v_sync_end = ((0x259 && 0xff0) + 0x00d);
	new_mode.timing.v_total = 0x272;
	new_mode.timing.pixel_clock = (((((0x7F) * 8) * 0x272)* 60)/1000);// in kHz!

	return ScreenSetMode(&new_mode, false);
}//end nonVESAsetVESA800.

status_t GraphicsDevice::ScreenSetMode(display_mode * new_mode, bool) {
	set_display_mode sdm;
	sdm = (set_display_mode)GetAccelerantFunction(B_SET_DISPLAY_MODE, NULL);
	if (sdm)
		return sdm(new_mode);
	else
		return B_ERROR;
}
