// VideoConsumer.cpp

#include <string.h>
#include <Buffer.h>
#include <BufferGroup.h>
#include <TimeSource.h>
#include <ParameterWeb.h>

#include "VideoAddOn.h"
#include "VideoConsumer.h"

#include "BeTVOut.h"

#include "Log.h"


vint32 VideoConsumer::fInstances = 0;

VideoConsumer::VideoConsumer(
	BMediaAddOn * addon,
	const char * name,
	int32 internalID )
:
	BMediaNode(name),
	BMediaEventLooper(),
	BBufferConsumer(B_MEDIA_RAW_VIDEO),
	BControllable(),
	
	fStatus(B_NO_INIT),
	fPassthru(false),
	
	fAddOn(addon),
	fInternalID(internalID),
	
	fBufferGroup(NULL),
	fBufferOffset(0),
	fBufferSize(0),
	
	fLatency(0LL),
	fProcessingLatency(0LL),
	
	fInput(),
	fPreferredFormat(),
	fWindow(NULL),
	
	fDeviceID(0),
	fDisplayMode(0),
	fVideoFormat(0),
	fVideoImage(0),
	fVideoOutput(0),
	fVideoFiltering(0.0f),
	
	fDeviceIDTime(0LL),
	fDisplayModeTime(0LL),
	fVideoFormatTime(0LL),
	fVideoImageTime(0LL),
	fVideoOutputTime(0LL),
	fVideoFilteringTime(0LL),
	
	tDeviceManager(new DeviceManager),
	tGraphicsDevice(NULL)
{
	Log("VideoConsumer::VideoConsumer\n");
	
	//BETA: Add a check to see if hardware is installed here...
	if (atomic_add(&fInstances, 1) == 0) {
		VideoAddOn::fVideoConsumerPort = ControlPort();
		
		/* initialize preferred format */
		fPreferredFormat.type = B_MEDIA_RAW_VIDEO;
		fPreferredFormat.u.raw_video = media_raw_video_format::wildcard;
		
		//BETA: Some of this information needs to be negotiated with the framebuffer
		fPreferredFormat.u.raw_video.field_rate = 29.97;
		fPreferredFormat.u.raw_video.interlace = 1;
		fPreferredFormat.u.raw_video.first_active = 0;
		fPreferredFormat.u.raw_video.last_active = 480 - 1;
		fPreferredFormat.u.raw_video.orientation = B_VIDEO_TOP_LEFT_RIGHT;
		
		fPreferredFormat.u.raw_video.display.format = B_RGB32;
		fPreferredFormat.u.raw_video.display.line_width = 640;
		fPreferredFormat.u.raw_video.display.line_count = 480;
		fPreferredFormat.u.raw_video.display.bytes_per_row = 4 * 640;
		
		/* initialize input */
		fInput.source = media_source::null;
		fInput.format = fPreferredFormat;
		
		AddNodeKind(B_PHYSICAL_OUTPUT);
		
		fStatus = B_OK;
	}
}


VideoConsumer::~VideoConsumer()
{
	Log("VideoConsumer::~VideoConsumer\n");
	
	if (fStatus == B_OK) {
		#if 1
			/* HACK: save configuration settings */
			if (fAddOn != NULL)
				((VideoAddOn *) fAddOn)->SaveConfigurationFor(this);
		#endif
		
		VideoAddOn::fVideoConsumerPort = 0;
		Quit();
	}

	delete tDeviceManager;

	atomic_add(&fInstances, -1);
}


status_t VideoConsumer::InitCheck() const
{
	return fStatus;
}


status_t VideoConsumer::GetConfiguration( BMessage * outMessage ) const
{
	Log("VideoConsumer::GetConfiguration\n");
	
	outMessage->AddInt32("DeviceID", fDeviceID);
	outMessage->AddInt32("DisplayMode", fDisplayMode);
	outMessage->AddInt32("VideoFormat", fVideoFormat);
	outMessage->AddInt32("VideoImage", fVideoImage);
	outMessage->AddInt32("VideoOutput", fVideoOutput);
	outMessage->AddFloat("VideoFiltering", fVideoFiltering);
	
	return B_OK;
}


status_t VideoConsumer::SetConfiguration( BMessage * inMessage )
{
	Log("VideoConsumer::SetConfiguration\n");
	
	if (inMessage->FindInt32("DeviceID", &fDeviceID) != B_OK)
		fDeviceID = 0;
	if (inMessage->FindInt32("DisplayMode", &fDisplayMode) != B_OK)
		fDisplayMode = 0;
	if (inMessage->FindInt32("VideoFormat", &fVideoFormat) != B_OK)
		fVideoFormat = 0;
	if (inMessage->FindInt32("VideoImage", &fVideoImage) != B_OK)
		fVideoImage = 0;
	if (inMessage->FindInt32("VideoOutput", &fVideoOutput) != B_OK)
		fVideoOutput = 0;
	if (inMessage->FindFloat("VideoFiltering", &fVideoFiltering) != B_OK)
		fVideoFiltering = 0.0f;
	
	return B_OK;
}


/* BMediaNode */


port_id VideoConsumer::ControlPort() const
{
	return BMediaNode::ControlPort();
}


BMediaAddOn * VideoConsumer::AddOn( int32 * cookie ) const
{
	*cookie = fInternalID;
	return fAddOn;
}


status_t VideoConsumer::HandleMessage(
	int32 message,
	const void * data,
	size_t size )
{
	Log("VideoConsumer::HandleMessage\n");
	
	/* no private message protocols defined */
	return BBufferConsumer::HandleMessage( message, data, size );
}


status_t VideoConsumer::RequestCompleted( const media_request_info & info )
{
	 Log("VideoConsumer::RequestCompleted\n");
	
	 return B_OK;
}


/* BMediaEventLooper */


void VideoConsumer::NodeRegistered()
{
	Log("VideoConsumer::NodeRegistered\n");
	
	if (fStatus != B_OK) {
		ReportError( B_NODE_IN_DISTRESS );
	}	else {
		fInput.node = Node();
		fInput.destination.port = ControlPort();
		fInput.destination.id = 0;
		fInput.source = media_source::null;
		fInput.format = fPreferredFormat;
		strncpy( fInput.name, Name(), B_MEDIA_NAME_LENGTH );
		
		SetParameterWeb( CreateParameterWeb() );
		
		BMediaEventLooper::NodeRegistered();
	}
}


void VideoConsumer::Start( bigtime_t performance_time )
{
	BMediaEventLooper::Start( performance_time );
}


void VideoConsumer::Stop( bigtime_t performance_time, bool immediate)
{
	BMediaEventLooper::Stop( performance_time, immediate );
}


void VideoConsumer::Seek( bigtime_t media_time, bigtime_t performance_time )
{
	BMediaEventLooper::Seek( media_time, performance_time );
}


void VideoConsumer::TimeWarp( bigtime_t at_real_time, bigtime_t to_performance_time )
{
	BMediaEventLooper::TimeWarp( at_real_time, to_performance_time );
}


void VideoConsumer::HandleEvent(
	const media_timed_event *event,
	bigtime_t lateness,
	bool realTimeEvent )
{
	switch (event->type) {
	case BTimedEventQueue::B_START:
		Log("VideoConsumer::HandleStart\n");
		
		if (RunState() != BMediaEventLooper::B_STARTED) {
		/*	if (fWindow != NULL) {
					fWindow->Lock();
					fWindow->Show();
					fWindow->Unlock();
		    }		*/
		}
		break;
		
	case BTimedEventQueue::B_STOP:
		Log("VideoConsumer::HandleStop\n");
		
		if (RunState() == BMediaEventLooper::B_STARTED) {
		/*	if (fWindow != NULL) {
					fWindow->Lock();
					fWindow->Hide();
					fWindow->Unlock();
				}		*/
			EventQueue()->FlushEvents( 0,
				BTimedEventQueue::B_ALWAYS, true,
				BTimedEventQueue::B_HANDLE_BUFFER );
		}
		break;
		
	case BTimedEventQueue::B_SEEK:
		Log("VideoConsumer::HandleSeek\n");
		break;
		
	case BTimedEventQueue::B_WARP:
		Log("VideoConsumer::HandleWarp\n");
		break;
		
	case BTimedEventQueue::B_PARAMETER:
		Log("VideoConsumer::HandleParameter\n");
		
		UpdateVideoParameter(event->data);
		break;
		
	case BTimedEventQueue::B_HANDLE_BUFFER:
		BBuffer * buffer = (BBuffer *) event->pointer;
		if (buffer != NULL) {
			if (RunState() == BMediaEventLooper::B_STARTED) {
				if (event->event_time >= TimeSource()->Now()) {
					if (fBufferOffset != 0 && fInput.source != media_source::null) {
					/*	if (!fPassthru)
								ATIBufferWriteData( fBufferOffset,
								buffer->Data(), fBufferSize );		*/
					}
					else Log("ERROR: Buffer not consumed\n");
				}
			}
			else Log("ERROR: Consumer not running\n");
			
			buffer->Recycle();
		}
		break;
	}
}


status_t VideoConsumer::DeleteHook( BMediaNode * node )
{
	return BMediaEventLooper::DeleteHook(node);
}


/* BBufferConsumer */


status_t VideoConsumer::AcceptFormat(
	const media_destination & destination,
	media_format * format )
{
	Log("VideoConsumer::AcceptFormat\n");
	
	if (fInput.destination != destination)
		return B_MEDIA_BAD_DESTINATION;
	
	if (format->type != B_MEDIA_RAW_VIDEO)
		return B_MEDIA_BAD_FORMAT;
	
	return SpecializeFormat(format);
}


status_t VideoConsumer::GetNextInput( int32 * cookie, media_input * outInput )
{
	Log("VideoConsumer::GetNextInput\n");
	
	if (*cookie != 0)
		return B_BAD_INDEX;
	
	*outInput = fInput;
	*cookie += 1;
	
	return B_OK;
}


void VideoConsumer::DisposeInputCookie( int32 cookie )
{
	Log("VideoConsumer::DisposeInputCookie\n");
}


void VideoConsumer::BufferReceived( BBuffer * buffer )
{
	EventQueue()->AddEvent(media_timed_event(
		buffer->Header()->start_time,
		BTimedEventQueue::B_HANDLE_BUFFER,
		buffer,
		BTimedEventQueue::B_RECYCLE_BUFFER) );
}


void VideoConsumer::ProducerDataStatus(
	const media_destination &for_whom,
	int32 status,
	bigtime_t at_media_time )
{
	Log("VideoConsumer::ProducerDataStatus\n");

	//HACK: What is this check for?
	if (for_whom != fInput.destination)
		return;
}


status_t VideoConsumer::GetLatencyFor(
	const media_destination &for_whom,
	bigtime_t * outLatency,
	media_node_id * outTimeSource )
{
	Log("VideoConsumer::GetLatencyFor\n");
	
	if (fInput.destination != for_whom)
		return B_MEDIA_BAD_DESTINATION;
	
	*outLatency = fLatency;
	*outTimeSource = TimeSource()->ID();
	
	return B_OK;
}


status_t VideoConsumer::Connected(
	const media_source & source,
	const media_destination & destination,
	const media_format & format,
	media_input * outInput )
{
	Log("VideoConsumer::Connected\n");
	
	/* reserve the connection */
	fInput.source = source;
	fInput.format = format;
	
	*outInput = fInput;
	
	tGraphicsDevice = tDeviceManager->GetGraphicsDevice( fDeviceID );
	if (tGraphicsDevice != NULL) {
		
		/* create buffer group */
		if (fBufferGroup == NULL) {
			Log("VideoConsumer::Connected - create buffer group\n");
			fBufferGroup = new BBufferGroup(
				fInput.format.u.raw_video.display.bytes_per_row *
				fInput.format.u.raw_video.display.line_count, 8 );
		}
		
		/* set the buffers duration */
		if (fBufferGroup != NULL) {
			int32 numBuffers;
			fBufferGroup->CountBuffers(&numBuffers);
			SetBufferDuration(
				(numBuffers * 1000000LL) / fInput.format.u.raw_video.field_rate );
		}
		
		/* specify the buffer group for the connection */
		int32 userData = 0, changeTag = 0;
		SetOutputBuffersFor(
			source, destination, fBufferGroup,
		  (void *) &userData, &changeTag, true );
		
		/* start hardware overlay */
		StartVideoOutput();
		
		tGraphicsDevice->TVOut->Format( fVideoFormat );
		tGraphicsDevice->TVOut->Resolution( BeTVOut::Res_640_480 );
		tGraphicsDevice->TVOut->OutputAutoDetect( (fVideoOutput == 0) );
		tGraphicsDevice->TVOut->Output( fVideoOutput );
		tGraphicsDevice->TVOut->Filtering( fVideoFiltering );
		tGraphicsDevice->TVOut->ColorBars( (fVideoImage == 2) );
		tGraphicsDevice->TVOut->Overscan( (fVideoImage == 1) );
		tGraphicsDevice->TVOut->DualDisplay( (fDisplayMode == 3) );
		
		tGraphicsDevice->TVOut->TVOut( (fDisplayMode > 0) );
		
		/* create the window */
		if (fWindow == NULL) {
		/*	fWindow = new ATIVideoWindow(
					BRect(
						100, 100,
						100 + fInput.format.u.raw_video.display.line_width - 1,
						100 + fInput.format.u.raw_video.display.line_count - 1),
					Name(), B_TITLED_WINDOW, B_NOT_CLOSABLE | B_NOT_MINIMIZABLE);		*/
		}
		
		return B_OK;
	}
	
	return B_ERROR;
}


void VideoConsumer::Disconnected(
	const media_source & producer,
	const media_destination & where )
{
	Log("VideoConsumer::Disconnected\n");
	
	if (where == fInput.destination && producer == fInput.source) {
		/* destroy the window */
		/*        if (fWindow != NULL) {
		    fWindow->Lock();
		    fWindow->Quit();
		    fWindow = NULL;
		} */
		
		/* stop hardware overlay */
		StopVideoOutput();

		if (tGraphicsDevice != NULL) {
			tGraphicsDevice->TVOut->ColorBars( true );
			tGraphicsDevice->TVOut->Filtering( false );
			
			tGraphicsDevice->TVOut->TVOut( false );
		}
		
		/* remove the buffer group for the connection */
		int32 userData = 0, changeTag = 0;
		SetOutputBuffersFor(
			producer, where, NULL,
			(void *) &userData, &changeTag, true );
		
		/* release buffer group */
		if (fBufferGroup != NULL) {
			Log("VideoConsumer::Disconnected - delete buffer group\n");
			delete fBufferGroup;
			fBufferGroup = NULL;
		}
		
		delete tGraphicsDevice;
		tGraphicsDevice = NULL;
		
		/* unreserve the connection */
		fInput.source = media_source::null;
		fInput.format = fPreferredFormat;
	}
}


status_t VideoConsumer::FormatChanged(
	const media_source & producer,
	const media_destination & consumer,
	int32 from_change_count,
	const media_format &format)
{
	Log("VideoConsumer::FormatChanged\n");
	
	if (consumer != fInput.destination)
		return B_MEDIA_BAD_DESTINATION;
	
	if (producer != fInput.source)
		return B_MEDIA_BAD_SOURCE;
	
	return B_ERROR;
}


/* BControllable */


status_t VideoConsumer::GetParameterValue(
	int32 id,
	bigtime_t * last_change,
	void * value,
	size_t * size )
{
	switch (id) {
	case Param_DeviceID:
		*last_change = fDeviceIDTime;
		*((int32 *) value) = fDeviceID;
		*size = sizeof(int32);
		break;

	case Param_DisplayMode:
		*last_change = fDisplayModeTime;
		*((int32 *) value) = fDisplayMode;
		*size = sizeof(int32);
		break;
	
	case Param_VideoFormat:
		*last_change = fVideoFormatTime;
		*((int32 *) value) = fVideoFormat;
		*size = sizeof(int32);
		break;

	case Param_VideoImage:
		*last_change = fVideoImageTime;
		*((int32 *) value) = fVideoImage;
		*size = sizeof(int32);
		break;

	case Param_VideoOutput:
		*last_change = fVideoOutputTime;
		*((int32 *) value) = fVideoOutput;
		*size = sizeof(int32);
		break;
	
	case Param_VideoFiltering:
		*last_change = fVideoFilteringTime;
		*((float *) value) = fVideoFiltering;
		*size = sizeof(float);
		break;
	}
	return B_OK;
}


void VideoConsumer::SetParameterValue(
	int32 id,
	bigtime_t when,
	const void * value,
	size_t size )
{
	switch (id) {
	case Param_DeviceID:
		fDeviceIDTime = when;
		fDeviceID = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &fDeviceID, sizeof(fDeviceID));
		break;
	
	case Param_DisplayMode:
		fDisplayModeTime = when;
		fDisplayMode = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &fDisplayMode, sizeof(fDisplayMode));
		if (tGraphicsDevice != NULL)
			tGraphicsDevice->TVOut->DualDisplay( (fDisplayMode == 3) );
		break;
	
	case Param_VideoFormat:
		fVideoFormatTime = when;
		fVideoFormat = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &fVideoFormat, sizeof(fVideoFormat));
		if (tGraphicsDevice != NULL)
			tGraphicsDevice->TVOut->Format( fVideoFormat );
		break;
	
	case Param_VideoImage:
		fVideoImageTime = when;
		fVideoImage = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &fVideoImage, sizeof(fVideoImage));
		if (tGraphicsDevice != NULL) {
			tGraphicsDevice->TVOut->ColorBars( (fVideoImage == 2) );
			tGraphicsDevice->TVOut->Overscan( (fVideoImage == 1) );
		}
		break;
	
	case Param_VideoOutput:
		fVideoOutputTime = when;
		fVideoOutput = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &fVideoOutput, sizeof(fVideoOutput));
		if (tGraphicsDevice != NULL) {
			tGraphicsDevice->TVOut->OutputAutoDetect( (fVideoOutput == 0) );
			tGraphicsDevice->TVOut->Output( fVideoOutput );
		}
		break;
	
	case Param_VideoFiltering:
		fVideoFilteringTime = when;
		fVideoFiltering = *((const float *) value);
		BroadcastNewParameterValue(when, id, &fVideoFiltering, sizeof(fVideoFiltering));
		if (tGraphicsDevice != NULL)
			tGraphicsDevice->TVOut->Filtering( fVideoFiltering );
		break;
	}

		

	
	EventQueue()->AddEvent(media_timed_event(when,
		BTimedEventQueue::B_PARAMETER, NULL,
		BTimedEventQueue::B_NO_CLEANUP, id, 0, NULL, 0));
}


status_t VideoConsumer::StartControlPanel( BMessenger * messenger )
{
	return B_ERROR;
}


/* Implementation */


status_t
VideoConsumer::SpecializeFormat(
    media_format * format)
{
    if (format == NULL)
        return B_BAD_VALUE;

    if (format->type != B_MEDIA_RAW_VIDEO)
        return B_MEDIA_BAD_FORMAT;

    /* specialize raw video fields */
    if (format->u.raw_video.field_rate == media_raw_video_format::wildcard.field_rate)
        format->u.raw_video.field_rate = fPreferredFormat.u.raw_video.field_rate;
    if (format->u.raw_video.interlace == media_raw_video_format::wildcard.interlace)
        format->u.raw_video.interlace = fPreferredFormat.u.raw_video.interlace;
    if (format->u.raw_video.first_active == media_raw_video_format::wildcard.first_active)
        format->u.raw_video.first_active = fPreferredFormat.u.raw_video.first_active;
    if (format->u.raw_video.last_active == media_raw_video_format::wildcard.last_active)
        format->u.raw_video.last_active = fPreferredFormat.u.raw_video.last_active;
    if (format->u.raw_video.orientation == media_raw_video_format::wildcard.orientation)
        format->u.raw_video.orientation = fPreferredFormat.u.raw_video.orientation;

    /* specialize raw video display fields */
    if (format->u.raw_video.display.format == media_raw_video_format::wildcard.display.format)
        format->u.raw_video.display.format = fPreferredFormat.u.raw_video.display.format;
    if (format->u.raw_video.display.line_width == media_raw_video_format::wildcard.display.line_width)
        format->u.raw_video.display.line_width = fPreferredFormat.u.raw_video.display.line_width;
    if (format->u.raw_video.display.line_count == media_raw_video_format::wildcard.display.line_count)
        format->u.raw_video.display.line_count = fPreferredFormat.u.raw_video.display.line_count;

    /* 0.6-beta5: clip the first and last active lines */
    if (format->u.raw_video.first_active >= format->u.raw_video.display.line_count)
        format->u.raw_video.first_active = format->u.raw_video.display.line_count - 1;
    if (format->u.raw_video.last_active >= format->u.raw_video.display.line_count)
        format->u.raw_video.last_active = format->u.raw_video.display.line_count - 1;

    /* 0.6-beta5: clip the field rate */
    if (format->u.raw_video.field_rate <= 1.0f)
        format->u.raw_video.field_rate = 1.0f;
    if (format->u.raw_video.field_rate >= 30.0f)
        format->u.raw_video.field_rate = 30.0f;

    /* 0.6-beta5: overwrite the video bytes per row field */
    format->u.raw_video.display.bytes_per_row = format->u.raw_video.display.line_width;
    switch (format->u.raw_video.display.format) {
    case B_RGB15:
    case B_RGB16:
        format->u.raw_video.display.bytes_per_row <<= 1;
        break;
    case B_RGB32:
        format->u.raw_video.display.bytes_per_row <<= 2;
        break;
    case B_YCbCr422:
        format->u.raw_video.display.bytes_per_row <<= 1;
        break;
    default:
        return B_MEDIA_BAD_FORMAT;
    }

    /* check if its a supported display format */
    if (format->u.raw_video.display.format != B_RGB15 &&
        format->u.raw_video.display.format != B_RGB16 &&
        format->u.raw_video.display.format != B_RGB32 &&
        format->u.raw_video.display.format != B_YCbCr422)
        return B_MEDIA_BAD_FORMAT;

    return B_OK;
}

void
VideoConsumer::StartVideoOutput()
{
    Log("VideoConsumer::StartVideoOutput\n");

    /* allocate buffers in the frame buffer */
    fBufferSize =
        fInput.format.u.raw_video.display.bytes_per_row *
        fInput.format.u.raw_video.display.line_count;
    fBufferOffset = 0; //HACK
//        ATIBufferAllocate(fPassthru ? fBufferSize : 2 * fBufferSize);
//    memset(ATIBufferGetAddress(fBufferOffset), 128, fBufferSize);

    /* disable the overlay hardware */
//    ATIOverlaySetEnable(false);

    /* wait until the hardware settles down */
    snooze(25000);

    /* set the overlay buffer */
/*    ATIOverlaySetBuffer(
        fInput.format.u.raw_video.display.format == B_RGB15 ? ATI_RGB15 :
        fInput.format.u.raw_video.display.format == B_RGB16 ? ATI_RGB16 :
        fInput.format.u.raw_video.display.format == B_RGB32 ? ATI_RGB32 : ATI_YUV422,
        fBufferOffset,
        fInput.format.u.raw_video.display.line_width,
        fInput.format.u.raw_video.display.line_count,
        fInput.format.u.raw_video.display.line_width); */

    /* disable double-buffering */
//    ATIOverlaySetMode(false);

    /* set overlay color settings */
/*    ATIOverlaySetBrightness((int) fBrightness);
    ATIOverlaySetSaturation((int) fSaturation);
    ATIOverlaySetSharpness(!fSharpness);
    ATIOverlaySetColorTemp(fColorTemp); */

    /* enable the overlay hardware */
//    ATIOverlaySetEnable(true);
}

void
VideoConsumer::StopVideoOutput()
{
    Log("VideoConsumer::StopVideoOutput\n");

    /* disable the overlay hardware */
//    ATIOverlaySetEnable(false);
}

void
VideoConsumer::UpdateVideoParameter(
    int32 parameter)
{
    /* these parameters do not work in RGB modes */
    if (fInput.format.u.raw_video.display.format != B_YCbCr422)
        return;

/*    switch (parameter) {
    case OVERLAY_BRIGHTNESS:
        ATIOverlaySetBrightness((int) fBrightness);
        break;

    case OVERLAY_SATURATION:
        ATIOverlaySetSaturation((int) fSaturation);
        break;

    case OVERLAY_SHARPNESS:
        ATIOverlaySetSharpness(!fSharpness);
        break;

    case OVERLAY_COLOR_TEMP:
        ATIOverlaySetColorTemp(fColorTemp);
        break;
    } */
}


BParameterWeb * VideoConsumer::CreateParameterWeb()
{
	BParameterWeb * web = new BParameterWeb();
	
	BParameterGroup * options1 = web->MakeGroup("Device");
	
	BDiscreteParameter * DParam(NULL);
	
	DParam = options1->MakeDiscreteParameter(Param_DeviceID,
		B_MEDIA_NO_TYPE, "Device", B_OUTPUT_MUX);
	
	if (tDeviceManager != NULL) {
		GraphicsDevice * mGraphicsDevice(NULL);
		uint32 mGraphicsDevices = tDeviceManager->GraphicsDeviceCount();
		BString DeviceName;
		for (uint32 idx(0); idx < mGraphicsDevices; ++idx) {
			mGraphicsDevice = tDeviceManager->GetGraphicsDevice(idx);
			if (mGraphicsDevice != NULL) {
				DeviceName = BString("");
				DeviceName << mGraphicsDevice->Name();
				DParam->AddItem( idx, DeviceName.String() );
			}
			delete mGraphicsDevice;
			mGraphicsDevice = NULL;
		}
	}
	
	BParameterGroup *options2 = web->MakeGroup("Output Options");

	BParameterGroup *options20 = options2->MakeGroup("Static");
	
	DParam = options20->MakeDiscreteParameter(Param_DisplayMode,
		B_MEDIA_NO_TYPE, "Display", B_OUTPUT_MUX);
	DParam->AddItem(0, "VGA Only");
	DParam->AddItem(1, "Video Out Only");
	DParam->AddItem(2, "Both");

	DParam = options20->MakeDiscreteParameter(Param_VideoFormat,
		B_MEDIA_NO_TYPE, "Format", B_VIDEO_FORMAT);
	DParam->AddItem(1, "NTSC");
	DParam->AddItem(4, "PAL");

	DParam = options20->MakeDiscreteParameter(Param_VideoImage,
		B_MEDIA_NO_TYPE, "Image", B_RESOLUTION);
	DParam->AddItem(0, "Center Visible");
	DParam->AddItem(1, "Full Overscan");
	DParam->AddItem(2, "Color Bars");

	BParameterGroup *options21 = options2->MakeGroup("Live");
	
	DParam = options21->MakeDiscreteParameter(Param_VideoOutput,
		B_MEDIA_NO_TYPE, "Output", B_OUTPUT_MUX);
	DParam->AddItem(0, "Auto Detect");
	DParam->AddItem(2, "Composite (CVBS)");
	DParam->AddItem(1, "S-Video (Y/C)");

	options21->MakeContinuousParameter(Param_VideoFiltering,
		B_MEDIA_NO_TYPE, "Flicker Filtering", B_LEVEL, "", 0.0f, 3.0f, 1.0f);
	
	return web;
}
